from typing import List, Tuple, Union

from vowpalwabbit import pyvw
from movado.mab_handler import MabHandler
import random


class MabHandlerCB(MabHandler):
    def __init__(
        self,
        arms: int,
        cover: int = 3,
        debug: bool = False,
        controller_params: dict = None,
        debug_path: str = "mab",
        skip_debug_initialization: bool = False,
    ):
        super().__init__(
            debug,
            controller_params=controller_params,
            debug_path=debug_path,
            skip_debug_initialization=skip_debug_initialization,
        )
        self._mab = pyvw.vw(
            "--cb_explore " + str(arms) + " --cover " + str(cover) + " --quiet"
        )

    def predict(
        self, context: List[float], probability: bool = False
    ) -> Union[int, Tuple[int, float]]:
        context_str: str = "| "
        for feature in context:
            context_str += str(feature) + " "
        context_str.strip()
        prediction: Tuple[int, float] = self.sample_probability_mass_function(
            self._mab.predict(context_str)
        )
        self._last_predict_probability = prediction[1]
        self._last_action = prediction[0]
        if probability:
            return (
                (prediction[0] + 1, self._last_predict_probability)
                if self._last_action == 0
                else (prediction[0] + 1, 1 - self._last_predict_probability)
            )
        return prediction[0] + 1

    @staticmethod
    def sample_probability_mass_function(
        probability_mass_function: List[float],
    ) -> Tuple[int, float]:
        total = sum(probability_mass_function)
        scale = 1 / total
        probability_mass_function = [x * scale for x in probability_mass_function]
        draw = random.random()
        sum_prob = 0.0
        for index, prob in enumerate(probability_mass_function):
            sum_prob += prob
            if sum_prob > draw:
                return index, prob



from typing import List

# noinspection PyUnresolvedReferences
from movado.mab_controller import MabController  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.voting_controller import VotingController  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.distance_controller import (
    DistanceController,
)  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.kernel_regression_model import (
    KernelRegressionModel,
)  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.hoeffding_adaptive_tree_model import (
    HoeffdingAdaptiveTreeModel,
)  # pylint: disable=unused-import

models: List[str] = [
    model.replace("Model", "")
    for model in globals().keys()
    if ("Model" in model) and len(model) > 5
]
controllers: List[str] = [
    ctrl.replace("Controller", "")
    for ctrl in globals().keys()
    if ("Controller" in ctrl) and len(ctrl) > 10
]



from typing import List, Union
from pathlib import Path
from abc import ABC, abstractmethod
import numpy as np


class MabHandler(ABC):
    def __init__(
        self,
        debug: bool = False,
        debug_path: str = "mab",
        skip_debug_initialization: bool = False,
        controller_params=None,
    ):
        self._mab = None
        self._last_predict_probability: float = -1
        self._last_action: Union[int, float] = -1
        self._debug = debug
        self._sample_prefix = ""
        self.__costs: List[float] = []
        self.__controller_params = controller_params
        self.__mab_debug = "movado_debug/" + debug_path + ".csv"
        if debug and not skip_debug_initialization:
            self.initialize_debug()

    def initialize_debug(self):
        Path("movado_debug").mkdir(exist_ok=True)
        Path(self.__mab_debug).open("w").close()
        if self.__controller_params:
            Path(self.__mab_debug).open("a").write(
                "Model_Parameters, Mean Cost, Action, Context, Probability\n"
            )
        else:
            Path(self.__mab_debug).open("a").write(
                "Mean Cost, Action, Context, Probability\n"
            )

    @abstractmethod
    def predict(self, context: List[float]) -> float:
        pass

    def learn(
        self,
        cost: float,
        context: List[float],
        forced_predict_probability: int = None,
        forced_action: Union[int, float] = None,
    ) -> None:
        sample: str = (
            self._sample_prefix
            + (str(self._last_action) if not forced_action else str(forced_action))
            + ":"
            + str(cost)
            + ":"
            + (
                str(self._last_predict_probability)
                if forced_predict_probability is None
                else str(forced_predict_probability)
            )
            + " | "
        )
        for feature in context:
            sample += str(feature) + " "
        self._mab.learn(sample)
        if self._debug:
            self.__costs.append(cost)
            optional_params = (
                (str(self.__controller_params) + ", ")
                if self.__controller_params
                else ""
            )
            Path(self.__mab_debug).open("a").write(
                optional_params
                + str(np.mean(self.__costs))
                + ", "
                + (
                    str(self._last_action)
                    if forced_action is None
                    else str(forced_action)
                )
                + ", "
                + str(context)
                + ", "
                + (
                    str(self._last_predict_probability)
                    if forced_predict_probability is None
                    else str(forced_predict_probability)
                )
                + "\n"
            )

    def get_mean_cost(self) -> float:
        if self.__costs:
            return float(np.mean(self.__costs))
        else:
            return 0

    def get_last_predict_probability(self) -> float:
        return self._last_predict_probability

    def set_last_predict_probability(self, last_predict_probability: float) -> None:
        self._last_predict_probability = last_predict_probability

    def get_last_action(self) -> Union[int, float]:
        return self._last_action



import abc
from abc import abstractmethod
from typing import List


class Estimator(abc.ABC):
    @abstractmethod
    def predict(self, X: List[float]) -> List[float]:
        pass

    @abstractmethod
    def train(self, X: List[float], y: List[float]) -> None:
        pass

    @abstractmethod
    def get_error(self) -> float:
        pass



from collections import OrderedDict
from typing import Union, Callable, List, Dict, Any, Optional, Tuple
from pathlib import Path

from movado.controller import Controller
from movado.estimator import Estimator
from movado.mab_handler import MabHandler
import numpy as np
import itertools
import random
import scipy.special

# This inputs are used to populate the symbol table for class retrieval
# noinspection PyUnresolvedReferences
from movado.mab_controller import MabController  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.distance_controller import (
    DistanceController,
)  # pylint: disable=unused-import


class VotingController(Controller):
    def __init__(
        self,
        controller,  # class reference => assumption of homogeneous controllers with different parameters
        estimator: Estimator,
        exact_fitness: Callable[[List[float]], List[float]],
        voters: int,
        params: "OrderedDict[str, List[Union[int, float, str]]]",
        self_exact: Optional[object] = None,
        is_point_in_context: bool = False,
        mab_weight: bool = True,
        debug=False,
    ):

        super(VotingController, self).__init__(
            estimator=estimator,
            exact_fitness=exact_fitness,
            debug=debug,
            self_exact=self_exact,
        )
        self.__is_point_in_context = is_point_in_context
        self.__debug = debug
        self.__is_soft: bool = False
        self.__controllers: List[Controller] = self.__get_controllers(
            params,
            controller,
            estimator,
            exact_fitness,
            voters,
            self_exact=self_exact,
            mab_weight=mab_weight,
            debug=debug,
        )
        self.__last_winners: List[Controller] = []
        self.__last_decision: int = -1
        self.initialize_debug()
        self.__params = params

    def initialize_debug(self):
        Path(self._controller_debug).open("a").write(
            "Majority_Model_Parameters, Majority_Size, Point, Exec_Time, Error, Mean Weight, Estimation\n"
        )

    def learn(
        self,
        is_exact: bool,
        point: List[float] = None,
        exec_time: float = None,
        mab: Optional[Tuple[MabHandler, Union[int, float]]] = None,
        mab_forced_probability: Optional[float] = None,
        mab_forced_action: Optional[Union[int, float]] = None,
        mab_weight: Optional[MabHandler] = None,
        mab_weight_forced_probability: Optional[float] = None,
        mab_weight_forced_action: Optional[Union[int, float]] = None,
        is_point_in_context: bool = True,
    ):
        for controller in self.__controllers:
            if controller in self.__last_winners:
                controller.learn(
                    is_exact=is_exact,
                    point=point,
                    exec_time=exec_time,
                    mab=(controller.get_mab(), self.__last_decision),
                    mab_forced_probability=None,
                    mab_weight=controller.get_weight_mab(),
                    mab_weight_forced_probability=None,
                    is_point_in_context=is_point_in_context,
                )
            else:
                controller.learn(
                    is_exact=is_exact,
                    point=point,
                    exec_time=exec_time,
                    mab=(controller.get_mab(), self.__last_decision),
                    mab_forced_probability=1,
                    mab_forced_action=(
                        float(
                            np.mean(
                                [
                                    winner.get_mab().get_last_action()
                                    for winner in self.__last_winners
                                ]
                            )
                        )
                        if not self.__is_soft
                        else self.__last_decision
                    ),
                    mab_weight=controller.get_weight_mab(),
                    mab_weight_forced_probability=None,
                    mab_weight_forced_action=None,
                    is_point_in_context=is_point_in_context,
                )

    def __get_controllers(
        self,
        params: "OrderedDict[str, List[Union[int, float, str]]]",
        controller,
        estimator: Estimator,
        exact_fitness: Callable[[List[float]], List[float]],
        voters: int,
        self_exact: Optional[object],
        mab_weight: bool,
        debug: bool,
    ) -> List[Controller]:
        # TODO random sample indices, not controllers to save memory
        controllers: List[Controller] = []
        controller_class = controller
        if not controller_class:
            raise Exception("Controller '" + str(controller) + "' was not found")
        random.seed(0)
        for current_vals in random.sample(
            list(itertools.product(*list(params.values()))),
            voters,
        ):
            current_params = {k: v for k, v in zip(params.keys(), current_vals)}
            controllers.append(
                controller_class(
                    estimator=estimator,
                    exact_fitness=exact_fitness,
                    self_exact=self_exact,
                    debug=debug,
                    skip_debug_initialization=True,
                    mab_weight=mab_weight,
                    **current_params
                )
            )
        if type(controllers[0]) is MabController:
            self.__is_soft = True
        mab: MabHandler = controllers[0].get_mab()
        mab_weight: MabHandler = controllers[0].get_weight_mab()
        if mab:
            mab.initialize_debug()
        if mab_weight:
            mab_weight.initialize_debug()
        return list(random.sample(controllers, voters))

    def __compute_weight_vector(self):
        return scipy.special.softmax(
            [-contr.get_mab().get_mean_cost() for contr in self.__controllers]
        )

    def compute_objective(
        self, point: List[int], decision_only: bool = False
    ) -> List[float]:
        decisions = []
        if self.__is_soft:
            for controller in self.__controllers:
                decisions.append(controller.compute_objective(point, probability=True))
            decision = 1 - np.average([d[1] for d in decisions])
        else:
            for controller in self.__controllers:
                decisions.append(
                    controller.compute_objective(point, decision_only=True)
                )
            decision = np.average(decisions, weights=self.__compute_weight_vector())
        if decision >= 0.5 or self._estimator.get_error() == 0.0:
            self.__last_decision = 1
            if self.__is_soft:
                self.__last_winners = [
                    controller_decision[0]
                    for controller_decision in zip(self.__controllers, decisions)
                    if controller_decision[1][0] == 1
                ]
            else:
                self.__last_winners = [
                    controller_decision[0]
                    for controller_decision in zip(self.__controllers, decisions)
                    if controller_decision[1] == 1
                ]
            out, exec_time = self._compute_exact(
                point, is_point_in_context=self.__is_point_in_context
            )
        else:
            self.__last_decision = 0
            if self.__is_soft:
                self.__last_winners = [
                    controller_decision[0]
                    for controller_decision in zip(self.__controllers, decisions)
                    if controller_decision[1][0] == 0
                ]
            else:
                self.__last_winners = [
                    controller_decision[0]
                    for controller_decision in zip(self.__controllers, decisions)
                    if controller_decision[1] == 0
                ]
            out, exec_time = self._compute_estimated(
                point,
                is_point_in_context=self.__is_point_in_context,
            )

        if self._debug:
            params = (
                self.__last_winners[0].get_parameters()
                if self.__last_winners
                else ["Model_Parameters", "Empty Hard Majority"]
            )
            self.write_debug(
                {
                    params[0]: params[1],
                    "Majority_Size": len(self.__last_winners),
                    "Point": point,
                    "Exec_Time": exec_time,
                    "Error": self._estimator.get_error(),
                    "Mean Weight": np.mean(
                        [
                            ctrl.get_weight_mab().get_last_action()
                            for ctrl in self.__controllers
                        ]
                    ),
                    "Estimation": int(not self.__last_decision),
                }
            )
        return out

    def write_debug(self, debug_info: Dict[str, Any]):
        Path(self._controller_debug).open("a").write(
            str(debug_info["Model_Parameters"])
            + ", "
            + str(debug_info["Majority_Size"])
            + ", "
            + str(debug_info["Point"])
            + ", "
            + str(debug_info["Exec_Time"])
            + ", "
            + str(debug_info["Error"])
            + ", "
            + str(debug_info["Mean Weight"])
            + ", "
            + str(debug_info["Estimation"])
            + "\n"
        )

    def get_mab(self):
        raise Exception("'get_mab': Unsupported method in voting controller")

    def get_weight_mab(self):
        raise Exception("'get_weight_mab': Unsupported method in voting controller")

    def get_parameters(self):
        return self.__params



from pathlib import Path
from typing import List, Callable, Dict, Any, Optional, Union, Tuple

from movado.controller import Controller
from movado.estimator import Estimator
from movado.mab_handler_cb import MabHandlerCB
from movado.mab_handler_cats import MabHandlerCATS


class MabController(Controller):
    def __init__(
        self,
        exact_fitness: Callable[[List[float]], List[float]],
        estimator: Estimator,
        self_exact: Optional[object] = None,
        debug: bool = False,
        skip_debug_initialization=False,
        cover: int = 3,
        mab_weight: bool = True,
        mab_weight_epsilon: float = 0.2,
        mab_weight_bandwidth: int = 1,
    ):
        super().__init__(
            exact_fitness=exact_fitness,
            estimator=estimator,
            self_exact=self_exact,
            debug=debug,
        )
        self.__params = (
            "Model_Parameters",
            {
                "cover": cover,
                "mab_weight_epsilon": mab_weight_epsilon,
                "mab_weight_bandwidth": mab_weight_bandwidth,
            },
        )

        self.__cover = cover
        self.__mab = MabHandlerCB(
            arms=2,
            debug=debug,
            cover=cover,
            controller_params={self.__params[0]: self.__params[1]},
        )
        self.__weight_mab = None
        if mab_weight:
            self.__weight_mab = MabHandlerCATS(
                debug=debug,
                epsilon=mab_weight_epsilon,
                bandwidth=mab_weight_bandwidth,
                controller_params={self.__params[0]: self.__params[1]},
                debug_path="mab_weight",
            )
        self.__is_first_call = True

        if self._debug and not skip_debug_initialization:
            self.initialize_debug()

    def initialize_debug(self):
        Path(self._controller_debug).open("a").write(
            "Model_Parameters, Point, Exec_Time, Error, Estimation\n"
        )

    def compute_objective(
        self,
        point: List[int],
        decision_only: bool = False,
        probability=False,
    ) -> Union[List[float], int, Tuple[int, float]]:
        decision = self.__mab.predict(
            self._compute_controller_context(point), probability=probability
        )
        accuracy = self._estimator.get_error()
        if probability:
            return (
                (decision[0] - 1, decision[1]) if accuracy != 0.0 else (1, decision[1])
            )
        if decision_only:
            return decision - 1
        if decision == 2 or accuracy == 0.0:
            out, exec_time = self._compute_exact(
                point,
                (self.__mab, 2),
                1 if accuracy == 0.0 else None,
                self.__weight_mab,
                1 if accuracy == 0.0 else None,
            )
        else:
            out, exec_time = self._compute_estimated(
                point, (self.__mab, 1), self.__weight_mab
            )

        # TODO probably this check can be done only once
        if self._debug:
            self.write_debug(
                {
                    "Model_Parameters": {
                        "epsilon": self.__cover,
                        "cover": self.__cover,
                    },
                    "Point": point,
                    "Exec_Time": exec_time,
                    "Error": self._estimator.get_error(),
                    "Estimation": 0 if decision == 2 or accuracy == 0.0 else 1,
                }
            )
        self.__is_first_call = False
        return out

    def write_debug(self, debug_info: Dict[str, Any]):
        Path(self._controller_debug).open("a").write(
            str(debug_info["Model_Parameters"])
            + ", "
            + str(debug_info["Point"])
            + ", "
            + str(debug_info["Exec_Time"])
            + ", "
            + str(debug_info["Error"])
            + ", "
            + str(debug_info["Estimation"])
            + "\n"
        )

    def get_mean_cost(self):
        return self.__mab.get_mean_cost()

    def get_mab(self) -> MabHandlerCB:
        return self.__mab

    def get_weight_mab(self) -> MabHandlerCATS:
        return self.__weight_mab

    def get_parameters(self):
        return self.__params



from pathlib import Path
from typing import Union, Optional, Dict, List, Any
import numpy as np
from sklearn.neighbors import KDTree, DistanceMetric

from movado.controller import Controller
from movado.mab_handler import MabHandler
from movado.mab_handler_cats import MabHandlerCATS

# TODO make lists global
from movado.mab_handler_cb import MabHandlerCB


class DistanceController(Controller):
    # do not annotate types in __init__, it throws "ABCMeta Object is not subscriptable"
    def __init__(
        self,
        exact_fitness,
        estimator,
        self_exact: Optional[object] = None,
        nth_nearest=3,
        distance_metric="minkowski",
        threshold="mab",
        debug=False,
        skip_debug_initialization=False,
        mab_epsilon: float = 0.2,
        mab_bandwidth: int = 1,
        mab_weight: bool = True,
        mab_weight_epsilon: float = 0.2,
        mab_weight_bandwidth: int = 1,
    ):
        super().__init__(
            exact_fitness=exact_fitness,
            estimator=estimator,
            self_exact=self_exact,
            debug=debug,
        )

        self.__evaluated_points: List[List[float]] = []
        self.__threshold: str = threshold
        self.__distance_metric = distance_metric
        self.__nth_nearest: int = nth_nearest
        self.__are_neighbours_to_recompute: bool = True
        self.__knn: Optional[KDTree] = None
        self.__distances: List[float] = []
        self.__mab = None
        self.__weight_mab = None
        self.__params = (
            "Model_Parameters",
            {
                "nth_nearest": self.__nth_nearest,
                "mab_epsilon": mab_epsilon,
                "mab_bandwidth": mab_bandwidth,
                "mab_weight_epsilon": mab_weight_epsilon,
                "mab_weight_bandwidth": mab_weight_bandwidth,
            },
        )
        if self.__threshold == "mab":
            self.__mab = MabHandlerCATS(
                debug=debug,
                epsilon=mab_epsilon,
                bandwidth=mab_bandwidth,
                controller_params={self.__params[0]: self.__params[1]},
                skip_debug_initialization=skip_debug_initialization,
            )
            if mab_weight:
                self.__weight_mab = MabHandlerCATS(
                    debug=debug,
                    epsilon=mab_weight_epsilon,
                    bandwidth=mab_weight_bandwidth,
                    debug_path="mab_weight",
                    controller_params={self.__params[0]: self.__params[1]},
                    skip_debug_initialization=skip_debug_initialization,
                )

        if self._debug and not skip_debug_initialization:
            self.initialize_debug()

    def initialize_debug(self):
        Path(self._controller_debug).open("a").write(
            "Model_Parameters, Threshold, Nth_Nearest_Distance, Point, Exec_Time, Error, Estimation\n"
        )

    def compute_objective(
        self, point: List[int], decision_only: bool = False
    ) -> Union[List[float], int]:
        out: List[float]
        mae: float
        self.__add_point(point)
        threshold = self.__get_threshold(point)
        nth_distance = self.__get_nth_nearest_distance(point)
        error = self._estimator.get_error()

        if nth_distance > threshold or error == 0.0:
            if decision_only:
                return 1
            out, exec_time = self._compute_exact(
                point,
                mab=(self.__mab, threshold * 100)
                if self.__threshold == "mab"
                else None,
                mab_forced_probability=1
                if self.__threshold == "mab" and error == 0.0
                else None,
                mab_weight=self.__weight_mab
                if self.__threshold == "mab" and self.__weight_mab
                else None,
                mab_weight_forced_probability=1
                if self.__threshold == "mab" and error == 0.0 and self.__weight_mab
                else None,
                is_point_in_context=False,
            )
        else:
            if decision_only:
                return 0
            out, exec_time = self._compute_estimated(
                point,
                mab=(self.__mab, threshold * 100)
                if self.__threshold == "mab"
                else None,
                mab_weight=self.__weight_mab
                if self.__threshold == "mab" and self.__weight_mab
                else None,
                is_point_in_context=False,
            )

        # TODO probably this check can be done only once
        if self._debug:
            self.write_debug(
                {
                    self.__params[0]: self.__params[1],
                    "Threshold": 0 if error == 0.0 else threshold,
                    "Nth_Nearest_Distance": 0
                    if not self.__evaluated_points
                    else nth_distance,
                    "Point": point,
                    "Exec_Time": exec_time,
                    "Error": error
                    if nth_distance > threshold or error == 0.0
                    else self._estimator.get_error(),
                    "Estimation": 0 if nth_distance > threshold or error == 0.0 else 1,
                }
            )
        return out

    def write_debug(self, debug_info: Dict[str, Any]):
        Path(self._controller_debug).open("a").write(
            str(debug_info["Model_Parameters"])
            + ", "
            + str(debug_info["Threshold"])
            + ", "
            + str(debug_info["Nth_Nearest_Distance"])
            + ", "
            + str(debug_info["Point"])
            + ", "
            + str(debug_info["Exec_Time"])
            + ", "
            + str(debug_info["Error"])
            + ", "
            + str(debug_info["Estimation"])
            + "\n"
        )

    def __add_point(self, point: List[float]) -> None:
        self.__evaluated_points.append(point)
        self.__are_neighbours_to_recompute = True

    def __get_threshold(self, point: List[int]) -> float:
        return getattr(self, "_get_threshold_" + self.__threshold)(point)

    def _get_threshold_mean(self, point):
        return np.mean(self._get_distances())

    def _get_threshold_median(self, point):
        return np.median(self._get_distances())

    def _get_threshold_mab(self, point) -> float:
        return self.__mab.predict(self._compute_controller_context(point)) / 100

    def _get_distances(self) -> np.ndarray:
        distances: np.ndarray = np.empty([0])
        for point in self.__evaluated_points:
            distances = np.append(distances, self.__get_nth_nearest_distance(point))
        return distances

    def __get_nth_nearest_distance(self, point: List[float]) -> float:
        if self.__are_neighbours_to_recompute:
            self.__knn = KDTree(
                self.__evaluated_points,
                metric=DistanceMetric.get_metric(self.__distance_metric),
            )
        k = min(self.__nth_nearest, len(self.__evaluated_points))
        distance = self.__knn.query([point], k=k)[0][0][k - 1]
        self.__distances.append(distance)
        diameter = max(self.__distances) - min(self.__distances)
        if diameter == 0:
            return 0
        else:
            return (distance - min(self.__distances)) / diameter

    def get_mean_cost(self) -> float:
        if self.__threshold != "mab":
            raise Exception("Mean cost is defined only for Controllers employing Mabs")
        return self.__mab.get_mean_cost()

    def get_mab(self) -> Optional[MabHandlerCATS]:
        if not self.__mab:
            return None
        return self.__mab

    def get_weight_mab(self) -> Optional[MabHandlerCATS]:
        if not self.__weight_mab:
            return None
        return self.__weight_mab

    def get_parameters(self):
        return self.__params



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['movado']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19,<2.0',
 'plotly>=4.14.3,<5.0.0',
 'river>=0.7.0,<0.8.0',
 'scikit-learn>=0.24.2,<0.25.0',
 'scipy>=1.5.4,<2.0.0',
 'vowpalwabbit>=8.9.0,<9.0.0']

setup_kwargs = {
    'name': 'movado',
    'version': '0.4.19',
    'description': 'Approximation utility for expensive fitness functions',
    'long_description': None,
    'author': 'Daniele Paletti',
    'author_email': 'danielepaletti98@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import river as rv

from movado.model import Model


class HoeffdingAdaptiveTreeModel(Model):
    def __init__(
        self,
        eta=2,
        budget=2000,
    ):
        super(HoeffdingAdaptiveTreeModel, self).__init__()
        self._model = rv.preprocessing.StandardScaler()
        self._model |= rv.feature_extraction.RBFSampler()

        self._model |= rv.tree.HoeffdingAdaptiveTreeRegressor(
            leaf_prediction="adaptive"
        )
        models = rv.utils.expand_param_grid(
            self._model,
            {
                "HoeffdingAdaptiveTreeRegressor": {
                    "model_selector_decay": [
                        0.35,
                        0.55,
                        0.75,
                        0.95,
                    ],
                },
                "RBFSampler": {"gamma": [1e-3, 1e-1, 1, 10]},
            },
        )
        self._model = rv.expert.SuccessiveHalvingRegressor(
            models=models,
            metric=rv.metrics.RMSE(),
            budget=budget,
            eta=eta,
            verbose=True,
        )
